﻿using System;
using System.Net;
using System.Threading;
using EventStore.ClientAPI;
using EventStore.ClientAPI.SystemData;

namespace EventStoreSubscriptionIssue
{
    class Program
    {
        static IEventStoreConnection _connection;

        static void Main(string[] args)
        {
            var sb = ConnectionSettings.Create();
            sb.SetDefaultUserCredentials(new UserCredentials("admin", "changeit"));
            _connection = EventStoreConnection.Create(sb.Build(), new IPEndPoint(IPAddress.Parse("127.0.0.1"), 1113));
            _connection.ConnectAsync().Wait();

            Subscribe();

            Console.WriteLine("Running. Press any key to exit...");
            Console.ReadKey(true);
        }

        private static volatile object syncLock = new object();
        private static Position? lastPosition;

        private static void Subscribe()
        {
            lock (syncLock)
            {
                Console.WriteLine($"{DateTime.UtcNow} - {Thread.CurrentThread.ManagedThreadId}: Subscribing to all events");
                _connection.SubscribeToAllFrom(lastPosition.GetValueOrDefault(Position.End), false, EventAppeared, null, SubscriptionDropped);
            }
        }

        private static void EventAppeared(EventStoreCatchUpSubscription subscription, ResolvedEvent resolvedEvent)
        {
            lock (syncLock)
            {
                if (lastPosition.GetValueOrDefault(Position.Start) >= resolvedEvent.OriginalPosition)
                {
                    // We've received an event that we've already dealt with. Stop this subscription.
                    Console.WriteLine($"{DateTime.UtcNow} - {Thread.CurrentThread.ManagedThreadId}: Received event {resolvedEvent.OriginalPosition} is before our checkpoint {lastPosition}. Stopping subscriber.");
                    subscription.Stop();
                }
                else
                {
                    Console.WriteLine($"{DateTime.UtcNow} - {Thread.CurrentThread.ManagedThreadId}: Received event {resolvedEvent.OriginalPosition} ({resolvedEvent.OriginalEvent.EventType}). Updating checkpoint.");
                    lastPosition = resolvedEvent.OriginalPosition;
                }
            }
        }

        private static void SubscriptionDropped(EventStoreCatchUpSubscription subscription, SubscriptionDropReason reason, Exception ex)
        {
            if (reason == SubscriptionDropReason.UserInitiated)
            {
                Console.WriteLine($"{DateTime.UtcNow} - {Thread.CurrentThread.ManagedThreadId}: Subscription dropped with reason {reason}. Hanging up our hat...");
            }
            else
            {
                Console.WriteLine($"{DateTime.UtcNow} - {Thread.CurrentThread.ManagedThreadId}: Subscription dropped with reason {reason}. Resubscribing...");
                Subscribe();
            }
        }
    }
}
